﻿using System;
using System.Collections.Generic;
using Analyzer.Core;

namespace Analyzer.Atoms
{
    public static partial class DivideTest
    {
        private class DividedAlignment
            : IEquatable<DividedAlignment>
        {
            private readonly SequenceInterval queryInterval;
            private readonly SequenceInterval targetInterval;
            private readonly Alignment originalAlignment;
            private readonly int hashCode;

            public DividedAlignment(Alignment alignment)
            {
                this.queryInterval = alignment.QueryInterval;
                this.targetInterval = alignment.TargetInterval;
                this.originalAlignment = alignment;
                this.hashCode = HashProvider.Hash(queryInterval, targetInterval, originalAlignment);
            }

            private DividedAlignment(SequenceInterval queryInterval, SequenceInterval targetInterval, Alignment originalAlignment)
            {
                this.queryInterval = queryInterval;
                this.targetInterval = targetInterval;
                this.originalAlignment = originalAlignment;
                this.hashCode = HashProvider.Hash(queryInterval, targetInterval, originalAlignment);
            }

            public static bool operator ==(DividedAlignment x, DividedAlignment y)
            {
                if ((object)x == null)
                {
                    return (object)y == null;
                }
                else
                {
                    return x.Equals(y);
                }
            }

            public static bool operator !=(DividedAlignment x, DividedAlignment y)
            {
                return !(x == y);
            }

            public List<DividedAlignment> Divide(List<SequencePosition> points)
            {
                List<SequencePosition> qPoints = FindAll(points, queryInterval.Start, queryInterval.End);
                List<SequencePosition> tPoints = FindAll(points, targetInterval.Start, targetInterval.End);
                if (originalAlignment.IsReversed)
                {
                    tPoints.Reverse();
                }
                int qi = 1;
                int ti = 1;
                SequencePosition qLast = qPoints[0];
                SequencePosition tLast = tPoints[0];
                List<DividedAlignment> res = new List<DividedAlignment>();
                while (qi < qPoints.Count || ti < tPoints.Count)
                {
                    SequencePosition qNew;
                    SequencePosition tNew;
                    if (qi == qPoints.Count)
                    {
                        qNew = MapBack(tPoints[ti]);
                        tNew = tPoints[ti];
                        ti += 1;
                    }
                    else if (ti == tPoints.Count)
                    {
                        qNew = qPoints[qi];
                        tNew = Map(qPoints[qi]);
                        qi += 1;
                    }
                    else
                    {
                        qNew = MapBack(tPoints[ti]);
                        tNew = Map(qPoints[qi]);
                        if (qNew.Position == qPoints[qi].Position)
                        {
                            tNew = tPoints[ti];
                            qi += 1;
                            ti += 1;
                        }
                        else if (tNew.Position == tPoints[ti].Position)
                        {
                            qNew = qPoints[qi];
                            qi += 1;
                            ti += 1;
                        }
                        else if (qNew.Position < qPoints[qi].Position)
                        {
                            tNew = tPoints[ti];
                            ti += 1;
                        }
                        else
                        {
                            qNew = qPoints[qi];
                            qi += 1;
                        }
                    }
                    if (qNew.Position > qLast.Position && tNew.Position > tLast.Position)
                    {
                        res.Add(new DividedAlignment(new SequenceInterval(qLast, qNew), new SequenceInterval(tLast, tNew), originalAlignment));
                    }
                    qLast = qNew;
                    tLast = tNew;
                }
                return res;
            }

            public override bool Equals(object obj)
            {
                return Equals(obj as DividedAlignment);
            }

            public bool Equals(DividedAlignment other)
            {
                if ((object)other == null)
                {
                    return false;
                }
                else
                {
                    return this.queryInterval == other.queryInterval && this.targetInterval == other.targetInterval && this.originalAlignment == other.originalAlignment;
                }
            }

            private static List<SequencePosition> FindAll(List<SequencePosition> points, SequencePosition start, SequencePosition end)
            {
                List<SequencePosition> res = new List<SequencePosition>();
                for (int i = FindFirst(points, start.Position); i < points.Count && points[i].Position <= end.Position; i++)
                {
                    res.Add(points[i]);
                }
                return res;
            }

            private static int FindFirst(List<SequencePosition> points, int position)
            {
                int x = 0;
                int y = points.Count;
                while (x < y)
                {
                    int m = (x + y) / 2;
                    if (points[m].Position >= position)
                    {
                        y = m;
                    }
                    else
                    {
                        x = m + 1;
                    }
                }
                return x;
            }

            private static int FindNearest(List<int> points, SequencePosition position)
            {
                int i = points.BinarySearch(position.Position);
                if (i < 0)
                {
                    i = ~i;
                }
                if (i == 0)
                {
                    return points[i];
                }
                else if (i == points.Count)
                {
                    return points[i - 1];
                }
                else if (position.Position - points[i - 1] < points[i] - position.Position)
                {
                    return points[i - 1];
                }
                else
                {
                    return points[i];
                }
            }

            public override int GetHashCode()
            {
                return hashCode;
            }

            public bool IsReversed
            {
                get
                {
                    return originalAlignment.IsReversed;
                }
            }

            private SequencePosition Map(SequencePosition position)
            {
                SequencePosition res = originalAlignment.Map(position, true);
                if (res.Position < targetInterval.Start.Position)
                {
                    return targetInterval.Start;
                }
                else if (res.Position > targetInterval.End.Position)
                {
                    return targetInterval.End;
                }
                else
                {
                    return res;
                }
            }

            private SequencePosition MapBack(SequencePosition position)
            {
                SequencePosition res = originalAlignment.MapBack(position, true);
                if (res.Position < queryInterval.Start.Position)
                {
                    return queryInterval.Start;
                }
                else if (res.Position > queryInterval.End.Position)
                {
                    return queryInterval.End;
                }
                else
                {
                    return res;
                }
            }

            public Alignment OriginalAlignment
            {
                get
                {
                    return originalAlignment;
                }
            }

            public SequenceInterval QueryInterval
            {
                get
                {
                    return queryInterval;
                }
            }

            public DividedAlignment SnapTo(List<int> points)
            {
                int qStart = FindNearest(points, queryInterval.Start);
                int qEnd = FindNearest(points, queryInterval.End);
                int tStart = FindNearest(points, targetInterval.Start);
                int tEnd = FindNearest(points, targetInterval.End);
                if (qStart < qEnd && tStart < tEnd)
                {
                    return new DividedAlignment(new SequenceInterval(queryInterval.Sequence, qStart, qEnd), new SequenceInterval(targetInterval.Sequence, tStart, tEnd), originalAlignment);
                }
                else
                {
                    return null;
                }
            }

            public SequenceInterval TargetInterval
            {
                get
                {
                    return targetInterval;
                }
            }
        }
    }
}
