﻿using System;
using System.Collections.Generic;
using Analyzer.Core;

namespace Analyzer.Atoms
{
    public static partial class DivideTest
    {
        private static void Atomize(List<DividedAlignment> alignments, SelfAlignmentSet s)
        {
            List<SequencePosition> endPoints = GetEndPoints(alignments);
            Console.Write("Done: {0} alignments, {1} atoms, ", alignments.Count, endPoints.Count - 1);
            Dictionary<SequenceInterval, AtomInfo> atoms = new Dictionary<SequenceInterval, AtomInfo>();
            foreach (DividedAlignment alignment in alignments)
            {
                if (!atoms.ContainsKey(alignment.QueryInterval))
                {
                    atoms.Add(alignment.QueryInterval, new AtomInfo(alignment.QueryInterval));
                }
                if (!atoms.ContainsKey(alignment.TargetInterval))
                {
                    atoms.Add(alignment.TargetInterval, new AtomInfo(alignment.TargetInterval));
                }
                atoms[alignment.QueryInterval].Edges.Add(atoms[alignment.TargetInterval]);
                atoms[alignment.TargetInterval].Edges.Add(atoms[alignment.QueryInterval]);
                atoms[alignment.QueryInterval].IsEdgeReversed.Add(alignment.IsReversed);
                atoms[alignment.TargetInterval].IsEdgeReversed.Add(alignment.IsReversed);
            }
            int classes = 0;
            int mispaired = 0;
            foreach (AtomInfo atom in atoms.Values)
            {
                if (atom.ClassNumber == 0)
                {
                    int count = 0;
                    int edges = 0;
                    classes += 1;
                    bool isMispaired = false;
                    AtomSearch(atom, classes, ref isMispaired, ref count, ref edges);
                    if (isMispaired)
                    {
                        mispaired += 1;
                    }
                    Console.WriteLine("Class {0} with {1} atoms and {2} edge usage factor.", atom.ClassNumber, count, (double)edges / (count * count));
                }
            }
            Console.WriteLine("{0} classes ({1} mispaired)", classes, mispaired);
            double sumAlignments = 0;
            double sumAtoms = 0;
            double sumIntersection = 0;
            foreach (Alignment a in s.Alignments)
            {
                if (a.QueryInterval.Start <= a.TargetInterval.Start)
                {
                    sumAlignments += a.QueryInterval.Length + a.TargetInterval.Length;
                }
            }
            foreach (AtomInfo a1 in atoms.Values)
            {
                foreach (AtomInfo a2 in atoms.Values)
                {
                    if (a1.ClassNumber == a2.ClassNumber || a1.ClassNumber == -a2.ClassNumber)
                    {
                        sumAtoms += a1.Interval.Length;
                        if (a1 == a2)
                        {
                            //sumAtoms += a2.Interval.Length;
                        }
                    }
                }
            }
            foreach (DividedAlignment a in alignments)
            {
                sumIntersection += a.QueryInterval.Intersect(a.OriginalAlignment.QueryInterval).Length + a.TargetInterval.Intersect(a.OriginalAlignment.TargetInterval).Length;
            }
            Console.WriteLine("{0:N2}% input adherence, {1:N2}% output consistency", 100 * sumIntersection / sumAlignments, 50 * sumIntersection / sumAtoms);
            SortedDictionary<int, int> lengths = new SortedDictionary<int, int>();
            foreach (SequenceInterval interval in atoms.Keys)
            {
                if (!lengths.ContainsKey(interval.Length))
                {
                    lengths[interval.Length] = 0;
                }
                lengths[interval.Length] += 1;
            }
            foreach (KeyValuePair<int, int> kvp in lengths)
            {
                Console.Write("{0} {1} ", kvp.Key, kvp.Value);
            }
            Console.WriteLine();
            Console.WriteLine("[Begin atomization]");
            int atomId = 0;
            foreach (AtomInfo a in atoms.Values)
            {
                Console.WriteLine("{0}\t{1}\t{2}\t{3}\t{4}\t{5}", s.Sequence.Name, atomId, Math.Abs(a.ClassNumber), Math.Sign(a.ClassNumber), a.Interval.Start.Position, a.Interval.End.Position);
                atomId += 1;
            }
            Console.WriteLine("[End atomization]");
        }

        private static void AtomSearch(AtomInfo atom, int num, ref bool isMispaired, ref int count, ref int edges)
        {
            if (atom.ClassNumber == 0)
            {
                atom.ClassNumber = num;
                count += 1;
                Console.Write("{0} {1} :", atom.ClassNumber, atom.Interval.Start.Position);
                for (int i = 0; i < atom.Edges.Count; i++)
                {
                    Console.Write(" {0}", atom.Edges[i].Interval.Start.Position * (atom.IsEdgeReversed[i] ? -1 : 1));
                }
                Console.WriteLine();
                edges -= 1;
                for (int i = 0; i < atom.Edges.Count; i++)
                {
                    edges += 1;
                    AtomInfo neighbor = atom.Edges[i];
                    if (atom.IsEdgeReversed[i])
                    {
                        AtomSearch(neighbor, -num, ref isMispaired, ref count, ref edges);
                    }
                    else
                    {
                        AtomSearch(neighbor, num, ref isMispaired, ref count, ref edges);
                    }
                }
            }
            else if (atom.ClassNumber != num)
            {
                isMispaired = true;
            }
        }

        private static List<int> Compactify(List<SequencePosition> points, int minDistance)
        {
            if (minDistance == 0)
            {
                List<int> res = new List<int>();
                foreach (SequencePosition point in points)
                {
                    res.Add(point.Position);
                }
                return res;
            }
            else
            {
                int length = points[points.Count - 1].Position;
                List<int> p = new List<int>();
                List<int> w = new List<int>();
                for (int i = 0; i < points.Count; i++)
                {
                    if (i == 0 || points[i].Position > points[i - 1].Position)
                    {
                        p.Add(points[i].Position);
                        w.Add(1);
                    }
                    else
                    {
                        w[w.Count - 1] += 1;
                    }
                }
                long[] best = new long[length + 1];
                int[] prev = new int[length + 1];
                best[0] = 0;
                prev[0] = 0;
                for (int i = 1; i < minDistance; i++)
                {
                    best[i] = int.MaxValue;
                    prev[i] = 0;
                }
                for (int i = minDistance; i <= length; i++)
                {
                    best[i] = int.MaxValue;
                    for (int j = i - minDistance; j > i - 2 * minDistance && j >= 0; j--)
                    {
                        long score = best[j];
                        int k = p.BinarySearch(j);
                        if (k < 0)
                        {
                            k = ~k;
                        }
                        while (k < points.Count && p[k] < i)
                        {
                            score += Math.Min(p[k] - j, i - p[k]) * w[k];
                            k += 1;
                        }
                        if (score < best[i])
                        {
                            best[i] = score;
                            prev[i] = j;
                        }
                    }
                }
                int pos = length;
                List<int> res = new List<int>();
                Sequence seq = points[0].Sequence;
                while (pos > 0)
                {
                    res.Add(pos);
                    pos = prev[pos];
                }
                res.Reverse();
                return res;
            }
        }

        private static bool Divide(ref List<DividedAlignment> alignments)
        {
            bool changes = false;
            List<SequencePosition> endPoints = GetEndPoints(alignments);
            endPoints.Sort();
            List<DividedAlignment> divided = new List<DividedAlignment>();
            foreach (DividedAlignment alignment in alignments)
            {
                List<DividedAlignment> div = alignment.Divide(endPoints);
                divided.AddRange(div);
                if (!changes)
                {
                    if (div.Count != 1 || div[0] != alignment)
                    {
                        changes = true;
                    }
                }
            }
            alignments = divided;
            return changes;
        }

        private static List<SequencePosition> GetEndPoints(List<DividedAlignment> alignments)
        {
            SortedDictionary<SequencePosition, object> dict = new SortedDictionary<SequencePosition, object>();
            foreach (DividedAlignment alignment in alignments)
            {
                dict[alignment.QueryInterval.Start] = null;
                dict[alignment.QueryInterval.End] = null;
                dict[alignment.TargetInterval.Start] = null;
                dict[alignment.TargetInterval.End] = null;
            }
            return new List<SequencePosition>(dict.Keys);
        }

        private static List<SequencePosition> GetEndPointsMulti(List<DividedAlignment> alignments)
        {
            List<SequencePosition> res = new List<SequencePosition>();
            foreach (DividedAlignment alignment in alignments)
            {
                res.Add(alignment.QueryInterval.Start);
                res.Add(alignment.QueryInterval.End);
                res.Add(alignment.TargetInterval.Start);
                res.Add(alignment.TargetInterval.End);
            }
            res.Sort();
            return res;
        }

        private static void Merge(ref List<DividedAlignment> alignments, int minDistance)
        {
            List<int> points = Compactify(GetEndPointsMulti(alignments), minDistance);
            List<DividedAlignment> res = new List<DividedAlignment>();
            foreach (DividedAlignment alignment in alignments)
            {
                DividedAlignment aligned = alignment.SnapTo(points);
                if (aligned != null)
                {
                    res.Add(aligned);
                }
            }
            alignments = res;
        }

        public static void Run(SelfAlignmentSet s, int minDistance)
        {
            Console.WriteLine("MinDistance = {0}", minDistance);
            List<DividedAlignment> alignments = new List<DividedAlignment>();
            foreach (Alignment alignment in s.Alignments)
            {
                if (alignment.QueryInterval.Start <= alignment.TargetInterval.Start)
                {
                    alignments.Add(new DividedAlignment(alignment));
                }
            }
            for (int step = 1; ; step++)
            {
                Console.WriteLine("Step {0}... ({1} alignments, {2} end points)", step, alignments.Count, GetEndPoints(alignments).Count);
                Merge(ref alignments, minDistance);
                if (!Divide(ref alignments))
                {
                    break;
                }
            }
            Atomize(alignments, s);
        }
    }
}
