﻿using System;
using System.Collections.Generic;

namespace Analyzer.Core
{
    public partial class Alignment
    {
        private readonly Sequence query;
        private readonly Sequence target;
        private readonly SequenceInterval queryInterval;
        private readonly SequenceInterval targetInterval;
        private readonly bool reversed;
        private readonly List<AlignmentBlock> blocks;
        private readonly List<AlignmentBlock> reverseBlocks;

        public Alignment(Sequence query, Sequence target, IEnumerable<AlignmentBlock> blocks)
        {
            this.query = query;
            this.target = target;
            foreach (AlignmentBlock block in blocks)
            {
                if (block.Query != query || block.Target != target)
                {
                    throw new ArgumentException("Blocks must align the query sequence to the target sequence.", "blocks");
                }
            }
            List<AlignmentBlock> blockList = new List<AlignmentBlock>(blocks);
            if (blockList.Count == 0)
            {
                throw new ArgumentException("You must specify at least one block.", "blocks");
            }
            blockList.Sort(CompareBlockStarts);
            for (int i = 1; i < blockList.Count; i++)
            {
                if (blockList[i - 1].End > blockList[i].Start)
                {
                    throw new ArgumentException("Blocks must not overlap.", "blocks");
                }
                if (blockList[i - 1].End != blockList[i].Start && blockList[i - 1].MappedEnd != blockList[i].MappedStart)
                {
                    throw new ArgumentException("Blocks must not have gaps on both query and target side at the same time.", "blocks");
                }
                if (blockList[i - 1].IsReversed != blockList[i].IsReversed)
                {
                    throw new ArgumentException("Blocks must all have the same orientation.", "blocks");
                }
            }
            this.reversed = blockList[0].IsReversed;
            this.queryInterval = new SequenceInterval(blockList[0].Start, blockList[blockList.Count - 1].End);
            this.blocks = new List<AlignmentBlock>();
            for (int i = 0; i < blockList.Count; i++)
            {
                if (i == 0 || i == blockList.Count - 1 || blockList[i - 1].End != blockList[i + 1].Start)
                {
                    this.blocks.Add(blockList[i]);
                }
            }
            blockList.Sort(CompareReverseBlockStarts);
            for (int i = 1; i < blockList.Count; i++)
            {
                if (blockList[i - 1].MappedEnd > blockList[i].MappedStart)
                {
                    throw new ArgumentException("Blocks must not overlap.", "blocks");
                }
            }
            this.targetInterval = new SequenceInterval(blockList[0].TargetInterval.Start, blockList[blockList.Count - 1].TargetInterval.End);
            this.reverseBlocks = new List<AlignmentBlock>();
            for (int i = 0; i < blockList.Count; i++)
            {
                if (i == 0 || i == blockList.Count - 1 || blockList[i - 1].TargetInterval.End != blockList[i + 1].TargetInterval.Start)
                {
                    this.reverseBlocks.Add(blockList[i]);
                }
            }
        }

        public IList<AlignmentBlock> Blocks
        {
            get
            {
                return blocks.AsReadOnly();
            }
        }

        private int CompareBlockStarts(AlignmentBlock x, AlignmentBlock y)
        {
            if (x.Start == y.Start)
            {
                return x.End - y.End;
            }
            else
            {
                return x.Start - y.Start;
            }
        }

        private int CompareReverseBlockStarts(AlignmentBlock x, AlignmentBlock y)
        {
            if (x.TargetInterval.Start == y.TargetInterval.Start)
            {
                return x.TargetInterval.End - y.TargetInterval.End;
            }
            else
            {
                return x.TargetInterval.Start - y.TargetInterval.Start;
            }
        }

        public SequencePosition End
        {
            get
            {
                return queryInterval.End;
            }
        }

        public bool IsReversed
        {
            get
            {
                return reversed;
            }
        }

        public SequencePosition Map(SequencePosition position)
        {
            if (!queryInterval.Contains(position))
            {
                throw new ArgumentOutOfRangeException("position");
            }
            int x = 0;
            int y = blocks.Count - 1;
            int pos = position.Position;
            while (x <= y)
            {
                int m = (x + y) / 2;
                if (blocks[m].Start.Position >= pos)
                {
                    y = m - 1;
                }
                else if (blocks[m].End.Position <= pos)
                {
                    x = m + 1;
                }
                else
                {
                    return blocks[m].Map(position);
                }
            }
            if (y < 0)
            {
                return blocks[x].MappedStart;
            }
            else if (x >= blocks.Count)
            {
                return blocks[y].MappedEnd;
            }
            else
            {
                return new SequencePosition(target, (blocks[y].MappedEnd.Position + blocks[x].MappedStart.Position) / 2);
            }
        }

        public SequencePosition Map(SequencePosition position, bool allowOutside)
        {
            if (!allowOutside)
            {
                return Map(position);
            }
            else
            {
                if (position.Position <= queryInterval.Start.Position)
                {
                    if (reversed)
                    {
                        return new SequencePosition(target, Math.Min(targetInterval.End.Position + (queryInterval.Start.Position - position.Position), target.Length));
                    }
                    else
                    {
                        return new SequencePosition(target, Math.Max(targetInterval.Start.Position - (queryInterval.Start.Position - position.Position), 0));
                    }
                }
                else if (position.Position >= queryInterval.End.Position)
                {
                    if (reversed)
                    {
                        return new SequencePosition(target, Math.Max(targetInterval.Start.Position - (position.Position - queryInterval.End.Position), 0));
                    }
                    else
                    {
                        return new SequencePosition(target, Math.Min(targetInterval.End.Position + (position.Position - queryInterval.End.Position), target.Length));
                    }
                }
                else
                {
                    return Map(position);
                }
            }
        }

        public SequencePosition MapBack(SequencePosition position)
        {
            if (!targetInterval.Contains(position))
            {
                throw new ArgumentOutOfRangeException("position");
            }
            int x = 0;
            int y = reverseBlocks.Count - 1;
            int pos = position.Position;
            while (x <= y)
            {
                int m = (x + y) / 2;
                if (reverseBlocks[m].TargetInterval.Start.Position >= pos)
                {
                    y = m - 1;
                }
                else if (reverseBlocks[m].TargetInterval.End.Position <= pos)
                {
                    x = m + 1;
                }
                else
                {
                    return reverseBlocks[m].MapBack(position);
                }
            }
            if (y < 0)
            {
                if (reverseBlocks[x].IsReversed)
                {
                    return reverseBlocks[x].End;
                }
                else
                {
                    return reverseBlocks[x].Start;
                }
            }
            else if (x >= reverseBlocks.Count)
            {
                if (reverseBlocks[y].IsReversed)
                {
                    return reverseBlocks[y].Start;
                }
                else
                {
                    return reverseBlocks[y].End;
                }
            }
            else
            {
                if (reverseBlocks[y].IsReversed)
                {
                    return new SequencePosition(target, (reverseBlocks[y].Start.Position + reverseBlocks[x].End.Position) / 2);
                }
                else
                {
                    return new SequencePosition(target, (reverseBlocks[y].End.Position + reverseBlocks[x].Start.Position) / 2);
                }
            }
        }

        public SequencePosition MapBack(SequencePosition position, bool allowOutside)
        {
            if (!allowOutside)
            {
                return MapBack(position);
            }
            else
            {
                if (position.Position <= targetInterval.Start.Position)
                {
                    if (reversed)
                    {
                        return new SequencePosition(target, Math.Min(queryInterval.End.Position + (targetInterval.Start.Position - position.Position), query.Length));
                    }
                    else
                    {
                        return new SequencePosition(target, Math.Max(queryInterval.Start.Position - (targetInterval.Start.Position - position.Position), 0));
                    }
                }
                else if (position.Position >= targetInterval.End.Position)
                {
                    if (reversed)
                    {
                        return new SequencePosition(target, Math.Max(queryInterval.Start.Position - (position.Position - targetInterval.End.Position), 0));
                    }
                    else
                    {
                        return new SequencePosition(target, Math.Min(queryInterval.End.Position + (position.Position - targetInterval.End.Position), query.Length));
                    }
                }
                else
                {
                    return MapBack(position);
                }
            }
        }

        public Sequence Query
        {
            get
            {
                return query;
            }
        }

        public SequenceInterval QueryInterval
        {
            get
            {
                return queryInterval;
            }
        }

        public SequencePosition Start
        {
            get
            {
                return queryInterval.Start;
            }
        }

        public Sequence Target
        {
            get
            {
                return target;
            }
        }

        public SequenceInterval TargetInterval
        {
            get
            {
                return targetInterval;
            }
        }
    }
}
