﻿using System;
using System.Collections.Generic;
using System.IO;

namespace Analyzer.Core
{
    public class MultiAlignmentSet
    {
        private readonly List<Sequence> queries;
        private readonly List<Sequence> targets;
        private readonly List<Alignment> alignments;

        private MultiAlignmentSet()
        {
            this.queries = new List<Sequence>();
            this.targets = new List<Sequence>();
            this.alignments = new List<Alignment>();
        }

        public MultiAlignmentSet(MultiAlignmentSet alignmentSet)
        {
            this.alignments = alignmentSet.alignments;
            this.queries = alignmentSet.queries;
            this.targets = alignmentSet.targets;
        }

        public IList<Alignment> Alignments
        {
            get
            {
                return alignments.AsReadOnly();
            }
        }

        public static MultiAlignmentSet FromPSL(Stream stream)
        {
            return FromPSL(stream, new Sequence[] { });
        }

        public static MultiAlignmentSet FromPSL(Stream stream, IEnumerable<Sequence> sequences)
        {
            using (StreamReader reader = new StreamReader(stream))
            {
                return FromPSL(reader, sequences);
            }
        }

        public static MultiAlignmentSet FromPSL(string path)
        {
            //if (File.Exists(Path.GetDirectoryName(path) + "\\" + Path.GetFileNameWithoutExtension(path) + ".fasta"))
            //{
            //    return FromPSL(path, Sequence.FromFASTA(Path.GetDirectoryName(path) + "\\" + Path.GetFileNameWithoutExtension(path) + ".fasta"));
            //}
            //else
            //{
                return FromPSL(path, new Sequence[] { });
            //}
        }

        public static MultiAlignmentSet FromPSL(string path, IEnumerable<Sequence> sequences)
        {
            using (StreamReader reader = new StreamReader(path))
            {
                return FromPSL(reader, sequences);
            }
        }

        public static MultiAlignmentSet FromPSL(TextReader reader)
        {
            return FromPSL(reader, new Sequence[] { });
        }

        public static MultiAlignmentSet FromPSL(TextReader reader, IEnumerable<Sequence> sequences)
        {
            MultiAlignmentSet res = new MultiAlignmentSet();
            Dictionary<string, Sequence> seqs = new Dictionary<string, Sequence>();
            string line;
            foreach (Sequence s in sequences)
            {
                seqs.Add(s.Name, s);
            }
            while ((line = reader.ReadLine()) != null)
            {
                string[] tokens = line.Split('\t');
                string strand = tokens[8];
                string qName = tokens[9];
                int qSize = int.Parse(tokens[10]);
                int qStart = int.Parse(tokens[11]);
                int qEnd = int.Parse(tokens[12]);
                string tName = tokens[13];
                int tSize = int.Parse(tokens[14]);
                int tStart = int.Parse(tokens[15]);
                int tEnd = int.Parse(tokens[16]);
                int blockCount = int.Parse(tokens[17]);
                int[] blockSizes = ToInts(tokens[18]);
                int[] qStarts = ToInts(tokens[19]);
                int[] tStarts = ToInts(tokens[20]);
                if (blockSizes.Length != blockCount || qStarts.Length != blockCount || tStarts.Length != blockCount)
                {
                    throw new ArgumentException();
                }
                Sequence query;
                Sequence target;
                List<AlignmentBlock> blocks = new List<AlignmentBlock>(blockCount);
                if (!seqs.TryGetValue(qName, out query))
                {
                    query = new Sequence(qSize, qName);
                    seqs.Add(qName, query);
                }
                if (!seqs.TryGetValue(tName, out target))
                {
                    target = new Sequence(tSize, tName);
                    seqs.Add(tName, target);
                }
                if (strand == "+")
                {
                    if (qStarts[0] != qStart || tStarts[0] != tStart)
                    {
                        blocks.Add(new AlignmentBlock(query, target, qStart, tStart, 0, false));
                    }
                    for (int i = 0; i < blockCount; i++)
                    {
                        blocks.Add(new AlignmentBlock(query, target, qStarts[i], tStarts[i], blockSizes[i], false));
                    }
                    if (qStarts[blockCount - 1] + blockSizes[blockCount - 1] != qEnd || tStarts[blockCount - 1] + blockSizes[blockCount - 1] != tEnd)
                    {
                        blocks.Add(new AlignmentBlock(query, target, qEnd, tEnd, 0, false));
                    }
                }
                else if (strand == "-")
                {
                    if (qSize - qStarts[blockCount - 1] - blockSizes[blockCount - 1] != qStart || tStarts[blockCount - 1] + blockSizes[blockCount - 1] != tEnd)
                    {
                        blocks.Add(new AlignmentBlock(query, target, qStart, tEnd, 0, true));
                    }
                    for (int i = blockCount - 1; i >= 0; i--)
                    {
                        blocks.Add(new AlignmentBlock(query, target, qSize - qStarts[i] - blockSizes[i], tStarts[i], blockSizes[i], true));
                    }
                    if (qSize - qStarts[0] != qEnd || tStarts[0] != tStart)
                    {
                        blocks.Add(new AlignmentBlock(query, target, qEnd, tStart, 0, true));
                    }
                }
                else
                {
                    throw new NotSupportedException("Strand '" + strand + "' is not supported.");
                }
                res.alignments.Add(new Alignment(query, target, blocks));
                if (!res.queries.Contains(query))
                {
                    res.queries.Add(query);
                }
                if (!res.targets.Contains(target))
                {
                    res.targets.Add(target);
                }
            }
            return res;
        }

        public SelfAlignmentSet Merge(int padding)
        {
            MultiAlignmentSet res = new MultiAlignmentSet();
            List<Sequence> seqs = new List<Sequence>();
            foreach (Sequence s in queries)
            {
                if (!seqs.Contains(s))
                {
                    seqs.Add(s);
                }
            }
            foreach (Sequence s in targets)
            {
                if (!seqs.Contains(s))
                {
                    seqs.Add(s);
                }
            }
            int length = 0;
            string name = "";
            foreach (Sequence s in seqs)
            {
                Console.WriteLine("{0} {1}", s.Name, s.Length);
                length += s.Length + padding;
                if (name != "")
                {
                    name += "+";
                }
                name += s.Name;
            }
            Sequence seq = new Sequence(length, name);
            res.queries.Add(seq);
            res.targets.Add(seq);
            foreach (Alignment a in alignments)
            {
                int qOffset = 0;
                int tOffset = 0;
                foreach (Sequence s in seqs)
                {
                    if (a.Query == s)
                    {
                        break;
                    }
                    qOffset += s.Length + padding;
                }
                foreach (Sequence s in seqs)
                {
                    if (a.Target == s)
                    {
                        break;
                    }
                    tOffset += s.Length + padding;
                }
                List<AlignmentBlock> blocks = new List<AlignmentBlock>();
                foreach (AlignmentBlock b in a.Blocks)
                {
                    blocks.Add(new AlignmentBlock(seq, seq, qOffset + b.QueryInterval.Start.Position, tOffset + b.TargetInterval.Start.Position, b.Length, b.IsReversed));
                }
                res.alignments.Add(new Alignment(seq, seq, blocks));
            }
            return new SelfAlignmentSet(res);
        }

        public IList<Sequence> Queries
        {
            get
            {
                return queries.AsReadOnly();
            }
        }

        public SelfAlignmentSet Subset(Sequence sequence)
        {
            if (!queries.Contains(sequence))
            {
                throw new ArgumentException("Sequence is not among query sequences in this set.");
            }
            if (!targets.Contains(sequence))
            {
                throw new ArgumentException("Sequence is not among target sequences in this set.");
            }
            MultiAlignmentSet res = new MultiAlignmentSet();
            foreach (Alignment a in alignments)
            {
                if (a.Query == sequence && a.Target == sequence)
                {
                    res.alignments.Add(a);
                }
            }
            res.queries.Add(sequence);
            res.targets.Add(sequence);
            return new SelfAlignmentSet(res);
        }

        public SelfAlignmentSet Subset(string sequenceName)
        {
            Sequence sequence = null;
            foreach (Sequence s in targets)
            {
                if (s.Name == sequenceName)
                {
                    sequence = s;
                }
            }
            return Subset(sequence);
        }

        public SingleAlignmentSet Subset(Sequence query, Sequence target)
        {
            if (!queries.Contains(query))
            {
                throw new ArgumentException("Query sequence was not found in this set.");
            }
            if (!targets.Contains(target))
            {
                throw new ArgumentException("Target sequence was not found in this set.");
            }
            MultiAlignmentSet res = new MultiAlignmentSet();
            foreach (Alignment a in alignments)
            {
                if (a.Query == query && a.Target == target)
                {
                    res.alignments.Add(a);
                }
            }
            res.queries.Add(query);
            res.targets.Add(target);
            return new SingleAlignmentSet(res);
        }

        public SingleAlignmentSet Subset(string targetName, string queryName)
        {
            Sequence target = null;
            Sequence query = null;
            foreach (Sequence s in targets)
            {
                if (s.Name == targetName)
                {
                    target = s;
                }
            }
            foreach (Sequence s in queries)
            {
                if (s.Name == queryName)
                {
                    query = s;
                }
            }
            return Subset(target, query);
        }

        public IList<Sequence> Targets
        {
            get
            {
                return targets.AsReadOnly();
            }
        }

        private static int[] ToInts(string str)
        {
            return ToInts(str.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries));
        }

        private static int[] ToInts(string[] strs)
        {
            int[] ints = new int[strs.Length];
            for (int i = 0; i < strs.Length; i++)
            {
                ints[i] = int.Parse(strs[i]);
            }
            return ints;
        }
    }
}
