﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Text;

namespace Analyzer.Core
{
    public sealed class Sequence
    {
        private readonly string name;
        private readonly Strand forwardStrand;
        private readonly Strand reverseStrand;
        private readonly ReadOnlyCollection<Strand> strands;
        private readonly Sequence masked;

        public Sequence(int length, string name)
            : this(new Strand(length), name)
        {
        }

        public Sequence(string data, string name)
            : this(new Strand(data), name)
        {
        }

        private Sequence(Strand forwardStrand, string name)
        {
            this.name = name;
            this.forwardStrand = forwardStrand;
            this.reverseStrand = forwardStrand.ReverseComplement;
            this.strands = new ReadOnlyCollection<Strand>(new List<Strand>(new Strand[] { this.forwardStrand, this.reverseStrand }));
            if (forwardStrand.IsMasked)
            {
                this.masked = this;
            }
            else
            {
                this.masked = new Sequence(forwardStrand.Masked, name);
            }
        }

        public char this[int i]
        {
            get
            {
                return forwardStrand[i];
            }
        }

        public Strand ForwardStrand
        {
            get
            {
                return forwardStrand;
            }
        }

        public static List<Sequence> FromFASTA(Stream stream)
        {
            using (StreamReader reader = new StreamReader(stream))
            {
                return FromFASTA(reader);
            }
        }

        public static List<Sequence> FromFASTA(string path)
        {
            using (StreamReader reader = new StreamReader(path))
            {
                return FromFASTA(reader);
            }
        }

        public static List<Sequence> FromFASTA(TextReader reader)
        {
            StringBuilder builder = null;
            string line;
            string name = null;
            List<Sequence> res = new List<Sequence>();
            while ((line = reader.ReadLine()) != null)
            {
                if (line.StartsWith(">"))
                {
                    if (builder != null)
                    {
                        res.Add(new Sequence(builder.ToString(), name));
                    }
                    builder = new StringBuilder();
                    name = line.Substring(1).Split()[0];
                }
                else if (builder == null)
                {
                    throw new ArgumentException("File is not in the correct format. (Must start with '>'.)");
                }
                else
                {
                    builder.Append(line);
                }
            }
            if (builder != null)
            {
                res.Add(new Sequence(builder.ToString(), name));
            }
            return res;
        }

        public bool IsMasked
        {
            get
            {
                return this == masked;
            }
        }

        public int Length
        {
            get
            {
                return forwardStrand.Length;
            }
        }

        public Sequence Masked
        {
            get
            {
                return masked;
            }
        }
        
        public string Name
        {
            get
            {
                return name;
            }
        }

        public Strand ReverseStrand
        {
            get
            {
                return reverseStrand;
            }
        }

        public IList<Strand> Strands
        {
            get
            {
                return strands;
            }
        }

        public SequenceInterval ToInterval()
        {
            return new SequenceInterval(this, 0, this.Length);
        }
    }
}
