﻿using System;

namespace Analyzer.Core
{
    public class SequenceInterval
        : IEquatable<SequenceInterval>
    {
        private readonly Sequence sequence;
        private readonly SequencePosition start;
        private readonly SequencePosition end;
        private readonly StrandInterval forwardInterval;
        private readonly int hashCode;

        public SequenceInterval(Sequence sequence, int start, int end)
            : this(new SequencePosition(sequence, start), new SequencePosition(sequence, end))
        {
        }

        public SequenceInterval(SequencePosition start, SequencePosition end)
        {
            if (start > end)
            {
                throw new ArgumentOutOfRangeException("end");
            }
            this.sequence = start.Sequence;
            this.start = start;
            this.end = end;
            this.forwardInterval = new StrandInterval(start.ToForwardStrand(), end.ToForwardStrand());
            this.hashCode = HashProvider.Hash(forwardInterval);
        }

        public SequenceInterval(Sequence sequence, StrandInterval forwardInterval)
        {
            this.sequence = sequence;
            this.start = new SequencePosition(sequence, forwardInterval.Start);
            this.end = new SequencePosition(sequence, forwardInterval.End);
            this.forwardInterval = forwardInterval;
            this.hashCode = HashProvider.Hash(forwardInterval);
        }

        public static bool operator ==(SequenceInterval x, SequenceInterval y)
        {
            if ((object)x == null)
            {
                return (object)y == null;
            }
            else
            {
                return x.Equals(y);
            }
        }

        public static bool operator !=(SequenceInterval x, SequenceInterval y)
        {
            return !(x == y);
        }

        public bool Contains(SequencePosition position)
        {
            return forwardInterval.Contains(position.ToForwardStrand());
        }

        public SequencePosition End
        {
            get
            {
                return end;
            }
        }

        public override bool Equals(object obj)
        {
            return Equals(obj as SequenceInterval);
        }

        public bool Equals(SequenceInterval other)
        {
            if ((object)other == null)
            {
                return false;
            }
            else
            {
                return this.forwardInterval == other.forwardInterval;
            }
        }

        public override int GetHashCode()
        {
            return hashCode;
        }

        public SequenceInterval Grow(int amount)
        {
            return new SequenceInterval(sequence, forwardInterval.Grow(amount));
        }

        public SequenceInterval Intersect(SequenceInterval interval)
        {
            if (!this.Intersects(interval))
            {
                throw new ArgumentException();
            }
            SequencePosition start;
            SequencePosition end;
            if (this.Start >= interval.Start)
            {
                start = this.Start;
            }
            else
            {
                start = interval.Start;
            }
            if (this.End <= interval.End)
            {
                end = this.End;
            }
            else
            {
                end = interval.End;
            }
            return new SequenceInterval(start, end);
        }

        public bool Intersects(SequenceInterval other)
        {
            return this.Sequence == other.Sequence && this.Start <= other.End && this.End >= other.Start;
        }

        public int Length
        {
            get
            {
                return forwardInterval.Length;
            }
        }

        public SequencePosition Start
        {
            get
            {
                return start;
            }
        }

        public Sequence Sequence
        {
            get
            {
                return sequence;
            }
        }

        public StrandInterval ToForwardStrand()
        {
            return forwardInterval;
        }

        public StrandInterval ToReverseStrand()
        {
            return forwardInterval.ReverseComplement;
        }
    }
}
