﻿using System;
using System.Text;

namespace Analyzer.Core
{
    public sealed class Strand
    {
        private readonly string data;
        private readonly Strand masked;
        private readonly Strand reverseComplement;

        public Strand(int length)
            : this(length, null)
        {
        }

        private Strand(int length, Strand reverseComplement)
            : this(new string('?', length), reverseComplement)
        {
        }

        public Strand(string data)
            : this(data, null)
        {
        }

        private Strand(string data, Strand reverseComplement)
        {
            if (data == null)
            {
                throw new ArgumentNullException("data");
            }
            this.data = data;
            if (reverseComplement == null)
            {
                StringBuilder maskedBuilder = new StringBuilder();
                StringBuilder reverseComplementBuilder = new StringBuilder();
                for (int i = 0; i < data.Length; i++)
                {
                    maskedBuilder.Append(Mask(data[i]));
                }
                for (int i = data.Length - 1; i >= 0; i--)
                {
                    reverseComplementBuilder.Append(Complement(data[i]));
                }
                if (maskedBuilder.ToString() == data)
                {
                    this.masked = this;
                }
                else
                {
                    this.masked = new Strand(maskedBuilder.ToString());
                }
                this.reverseComplement = new Strand(reverseComplementBuilder.ToString(), this);
            }
            else
            {
                if (reverseComplement.IsMasked)
                {
                    this.masked = this;
                }
                else
                {
                    this.masked = reverseComplement.Masked.ReverseComplement;
                }
                this.reverseComplement = reverseComplement;
            }
        }

        public char this[int i]
        {
            get
            {
                return data[i];
            }
        }

        private static char Complement(char c)
        {
            switch (c)
            {
                case 'A':
                    return 'T';
                case 'C':
                    return 'G';
                case 'G':
                    return 'C';
                case 'T':
                    return 'A';
                case 'a':
                    return 't';
                case 'c':
                    return 'g';
                case 'g':
                    return 'c';
                case 't':
                    return 'a';
                default:
                    return c;
            }
        }

        public string Data
        {
            get
            {
                return data;
            }
        }

        public bool IsMasked
        {
            get
            {
                return this == masked;
            }
        }

        public int Length
        {
            get
            {
                return data.Length;
            }
        }

        private static string Mask(char c)
        {
            switch (c)
            {
                case 'A':
                    return "A";
                case 'C':
                    return "C";
                case 'G':
                    return "G";
                case 'T':
                    return "T";
                default:
                    return "";
            }
        }

        public Strand Masked
        {
            get
            {
                return masked;
            }
        }

        public Strand ReverseComplement
        {
            get
            {
                return reverseComplement;
            }
        }

        public StrandInterval ToInterval()
        {
            return new StrandInterval(this, 0, this.Length);
        }
    }
}
