﻿using System;

namespace Analyzer.Core
{
    public sealed class StrandInterval
        : IEquatable<StrandInterval>
    {
        private readonly StrandPosition start;
        private readonly StrandPosition end;
        private readonly StrandInterval reverseComplement;
        private readonly int hashCode;

        public StrandInterval(Strand strand, int start, int end)
            : this(new StrandPosition(strand, start), new StrandPosition(strand, end))
        {
        }

        public StrandInterval(StrandPosition start, StrandPosition end)
        {
            if (start > end)
            {
                throw new ArgumentOutOfRangeException("end");
            }
            this.start = start;
            this.end = end;
            this.reverseComplement = new StrandInterval(this);
            this.hashCode = HashProvider.Hash(start, end);
        }

        private StrandInterval(StrandInterval reverseComplement)
        {
            this.start = reverseComplement.end.ReverseComplement;
            this.end = reverseComplement.start.ReverseComplement;
            this.reverseComplement = reverseComplement;
            this.hashCode = HashProvider.Hash(start, end);
        }

        public static bool operator ==(StrandInterval x, StrandInterval y)
        {
            if ((object)x == null)
            {
                return (object)y == null;
            }
            else
            {
                return x.Equals(y);
            }
        }

        public static bool operator !=(StrandInterval x, StrandInterval y)
        {
            return !(x == y);
        }

        public bool Contains(StrandPosition position)
        {
            return position.Strand == this.Strand && position.Position >= this.start.Position && position.Position <= this.end.Position;
        }

        public StrandPosition End
        {
            get
            {
                return end;
            }
        }

        public override bool Equals(object obj)
        {
            return Equals(obj as StrandInterval);
        }

        public bool Equals(StrandInterval other)
        {
            if ((object)other == null)
            {
                return false;
            }
            else
            {
                return this.start == other.start && this.end == other.end;
            }
        }

        public override int GetHashCode()
        {
            return hashCode;
        }

        public StrandInterval Grow(int amount)
        {
            int s = Math.Max(start.Position - amount, 0);
            int e = Math.Min(end.Position + amount, this.Strand.Length);
            if (e < s)
            {
                e = (e + s) / 2;
                s = e;
            }
            return new StrandInterval(this.Strand, s, e);
        }

        public int Length
        {
            get
            {
                return end - start;
            }
        }

        public StrandInterval ReverseComplement
        {
            get
            {
                return reverseComplement;
            }
        }

        public StrandPosition Start
        {
            get
            {
                return start;
            }
        }

        public Strand Strand
        {
            get
            {
                return start.Strand;
            }
        }
    }
}
