﻿using System;

namespace Analyzer.Core
{
    public sealed class StrandPosition
        : IComparable<StrandPosition>, IEquatable<StrandPosition>
    {
        private readonly Strand strand;
        private readonly int position;
        private readonly StrandPosition reverseComplement;
        private readonly int hashCode;

        public StrandPosition(Strand strand, int position)
        {
            if (position < 0 || position > strand.Length)
            {
                throw new ArgumentOutOfRangeException("position");
            }
            this.strand = strand;
            this.position = position;
            this.reverseComplement = new StrandPosition(this);
            this.hashCode = HashProvider.Hash(strand, position);
        }

        private StrandPosition(StrandPosition reverseComplement)
        {
            this.strand = reverseComplement.strand.ReverseComplement;
            this.position = reverseComplement.strand.Length - reverseComplement.position;
            this.reverseComplement = reverseComplement;
            this.hashCode = HashProvider.Hash(strand, position);
        }

        public static bool operator ==(StrandPosition x, StrandPosition y)
        {
            if ((object)x == null)
            {
                return (object)y == null;
            }
            else
            {
                return x.Equals(y);
            }
        }

        public static bool operator !=(StrandPosition x, StrandPosition y)
        {
            return !(x == y);
        }

        public static bool operator <(StrandPosition x, StrandPosition y)
        {
            return x.CompareTo(y) < 0;
        }

        public static bool operator >(StrandPosition x, StrandPosition y)
        {
            return x.CompareTo(y) > 0;
        }

        public static bool operator <=(StrandPosition x, StrandPosition y)
        {
            return x.CompareTo(y) <= 0;
        }

        public static bool operator >=(StrandPosition x, StrandPosition y)
        {
            return x.CompareTo(y) >= 0;
        }

        public static StrandPosition operator +(StrandPosition x, int y)
        {
            return new StrandPosition(x.strand, x.position + y);
        }

        public static StrandPosition operator -(StrandPosition x, int y)
        {
            return new StrandPosition(x.strand, x.position - y);
        }

        public static int operator -(StrandPosition x, StrandPosition y)
        {
            if (x.strand != y.strand)
            {
                throw new ArgumentException("Positions are not on the same strand.");
            }
            else
            {
                return x.position - y.position;
            }
        }

        public int CompareTo(StrandPosition other)
        {
            return this - other;
        }

        public override bool Equals(object obj)
        {
            return Equals(obj as StrandPosition);
        }

        public bool Equals(StrandPosition other)
        {
            if ((object)other == null)
            {
                return false;
            }
            else
            {
                return this.strand == other.strand && this.position == other.position;
            }
        }

        public override int GetHashCode()
        {
            return hashCode;
        }

        public int Position
        {
            get
            {
                return position;
            }
        }

        public StrandPosition ReverseComplement
        {
            get
            {
                return reverseComplement;
            }
        }

        public Strand Strand
        {
            get
            {
                return strand;
            }
        }
    }
}
