﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using Analyzer.Core;

namespace Analyzer.Plot
{
    public static class Plotter
    {
        private static float AxisLength(AxisScale scale, int minValue, int maxValue)
        {
            switch (scale)
            {
                case AxisScale.Linear:
                    return maxValue - minValue;
                case AxisScale.Logarithmic:
                    int res = 0;
                    if (minValue == 0)
                    {
                        minValue = 1;
                        res += 1;
                    }
                    while (minValue < maxValue)
                    {
                        minValue *= 2;
                        res += 1;
                    }
                    return res;
                default:
                    throw new NotSupportedException();
            }
        }

        private static void DrawAxes(Graphics g, AxisScale xScale, AxisScale yScale, int xMinValue, int xMaxValue, int yMinValue, int yMaxValue, int imageWidth, int imageHeight)
        {
            DrawXAxis(g, 80, imageHeight - 30, imageWidth - 120, 5, xScale, xMinValue, xMaxValue, false);
            DrawYAxis(g, 80, imageHeight - 30, 50 - imageHeight, 5, yScale, yMinValue, yMaxValue, true);
            g.TranslateTransform(80, imageHeight - 30);
            g.ScaleTransform((float)(imageWidth - 120) / AxisLength(xScale, xMinValue, xMaxValue), (float)(50 - imageHeight) / AxisLength(yScale, yMinValue, yMaxValue));
        }

        private static void DrawXAxis(Graphics g, float x0, float y0, float length, float axisPadding, AxisScale scale, int minValue, int maxValue, bool hideFirstPoint)
        {
            DrawXAxis(g, x0, y0, length, axisPadding, GetCoordinates(scale, x0, x0 + length, minValue, maxValue, hideFirstPoint), GetValues(scale, minValue, maxValue, hideFirstPoint));
        }

        private static void DrawXAxis(Graphics g, float x0, float y0, float length, float axisPadding, IList<float> coordinates, IList<int> values)
        {
            g.DrawLine(Pens.Black, x0, y0, x0 + length, y0);
            Font f = new Font(FontFamily.GenericSansSerif, 12);
            for (int i = 0; i < coordinates.Count; i++)
            {
                g.DrawLine(Pens.Black, coordinates[i], y0, coordinates[i], y0 + axisPadding);
                g.DrawString(values[i].ToString(), f, Brushes.Black, coordinates[i] - g.MeasureString(values[i].ToString(), f).Width / 2, y0 + axisPadding);
            }
        }

        private static void DrawYAxis(Graphics g, float x0, float y0, float length, float axisPadding, AxisScale scale, int minValue, int maxValue, bool hideFirstPoint)
        {
            DrawYAxis(g, x0, y0, length, axisPadding, GetCoordinates(scale, y0, y0 + length, minValue, maxValue, hideFirstPoint), GetValues(scale, minValue, maxValue, hideFirstPoint));
        }

        private static void DrawYAxis(Graphics g, float x0, float y0, float length, float axisPadding, IList<float> coordinates, IList<int> values)
        {
            g.DrawLine(Pens.Black, x0, y0, x0, y0 + length);
            Font f = new Font(FontFamily.GenericSansSerif, 12);
            for (int i = 0; i < coordinates.Count; i++)
            {
                g.DrawLine(Pens.Black, x0, coordinates[i], x0 - axisPadding, coordinates[i]);
                g.DrawString(values[i].ToString(), f, Brushes.Black, x0 - axisPadding - g.MeasureString(values[i].ToString(), f).Width, coordinates[i] - g.MeasureString(values[i].ToString(), f).Height / 2);
            }
        }

        private static IList<float> GetCoordinates(AxisScale scale, float c0, float c1, int minValue, int maxValue, bool hideFirst)
        {
            switch (scale)
            {
                case AxisScale.Linear:
                    if (hideFirst)
                    {
                        return new float[] { c1 };
                    }
                    else
                    {
                        return new float[] { c0, c1 };
                    }
                case AxisScale.Logarithmic:
                    int count = 1;
                    if (minValue == 0)
                    {
                        minValue = 1;
                        count += 1;
                    }
                    while (minValue < maxValue)
                    {
                        minValue *= 2;
                        count += 1;
                    }
                    List<float> coords = new List<float>();
                    for (int i = (hideFirst ? 1 : 0); i < count; i++)
                    {
                        coords.Add(c0 + (c1 - c0) * i / (count - 1));
                    }
                    return coords;
                default:
                    throw new NotSupportedException();
            }
        }

        private static IList<int> GetValues(AxisScale scale, int minValue, int maxValue, bool hideFirst)
        {
            switch (scale)
            {
                case AxisScale.Linear:
                    if (hideFirst)
                    {
                        return new int[] { maxValue };
                    }
                    else
                    {
                        return new int[] { minValue, maxValue };
                    }
                case AxisScale.Logarithmic:
                    List<int> values = new List<int>();
                    if (!hideFirst)
                    {
                        values.Add(minValue);
                    }
                    if (minValue == 0)
                    {
                        minValue = 1;
                        values.Add(minValue);
                    }
                    while (minValue < maxValue)
                    {
                        minValue *= 2;
                        values.Add(minValue);
                    }
                    return values;
                default:
                    throw new NotSupportedException();
            }
        }

        public static Image PlotAlignment(SingleAlignmentSet alignments, int width)
        {
            Bitmap bmp = new Bitmap(width, width * alignments.Query.Length / alignments.Target.Length);
            Graphics g = Graphics.FromImage(bmp);
            g.SmoothingMode = SmoothingMode.HighQuality;
            g.TranslateTransform(0, bmp.Height);
            g.ScaleTransform((float)bmp.Width / alignments.Target.Length, -(float)bmp.Height / alignments.Query.Length);
            foreach (Alignment align in alignments.Alignments)
            {
                foreach (AlignmentBlock block in align.Blocks)
                {
                    g.DrawLine(Pens.Black, block.MappedStart.Position, block.Start.Position, block.MappedEnd.Position, block.End.Position);
                }
            }
            return bmp;
        }

        public static Image PlotData(IEnumerable<int> data, DataMode dataMode, AxisScale xScale, AxisScale yScale, int imageWidth, int imageHeight)
        {
            return PlotData(new IEnumerable<int>[] { data }, dataMode, xScale, yScale, imageWidth, imageHeight);
        }

        public static Image PlotData<T>(IEnumerable<T> data, DataMode dataMode, AxisScale xScale, AxisScale yScale, int imageWidth, int imageHeight) where T : IEnumerable<int>
        {
            Bitmap bmp = new Bitmap(imageWidth, imageHeight);
            Graphics g = Graphics.FromImage(bmp);
            g.SmoothingMode = SmoothingMode.HighQuality;
            List<List<int>> values = new List<List<int>>();
            int xMax = 0;
            int yMax = 0;
            foreach (IEnumerable<int> dataSet in data)
            {
                List<int> l = new List<int>(dataSet);
                l.Sort();
                values.Add(l);
                xMax = Math.Max(xMax, l[l.Count - 1]);
                yMax = Math.Max(yMax, l.Count);
            }
            DrawAxes(g, xScale, yScale, 0, xMax, 0, yMax, imageWidth, imageHeight);
            Random rand = new Random(0);
            foreach (List<int> valueList in values)
            {
                PlotData(g, Color.FromArgb(rand.Next(256), rand.Next(256), rand.Next(256)), valueList, dataMode, xScale, yScale);
            }
            return bmp;
        }

        private static void PlotData(Graphics g, Color c, List<int> data, DataMode mode, AxisScale xScale, AxisScale yScale)
        {
            Pen pen = new Pen(c, 0);
            PointF p = new PointF(0, 0);
            switch (mode)
            {
                case DataMode.Averaged:
                    float window = (float)Transform(data[data.Count - 1], xScale) / 10;
                    int i0 = 0;
                    int i1 = 0;
                    while (i0 < data.Count)
                    {
                        float x0 = Transform(data[i0], xScale);
                        float x1 = Transform(data[i1], xScale);
                        PointF p2 = new PointF((x0 + x1) / 2, p.Y);
                        PointF p3 = new PointF((x0 + x1) / 2, Transform(i1 - i0, yScale));
                        g.DrawLine(pen, p, p2);
                        g.DrawLine(pen, p2, p3);
                        p = p3;
                        if (i1 + 1 < data.Count && Transform(data[i1 + 1], xScale) - x0 <= window)
                        {
                            i1 += 1;
                            while (i1 + 1 < data.Count && data[i1 + 1] == data[i1])
                            {
                                i1 += 1;
                            }
                        }
                        else
                        {
                            i0 += 1;
                            while (i0 + 1 < data.Count && data[i0 + 1] == data[i0])
                            {
                                i0 += 1;
                            }
                        }
                    }
                    break;
                case DataMode.Cumulative:
                    for (int i = 0; i < data.Count; i++)
                    {
                        while (i + 1 < data.Count && data[i + 1] == data[i])
                        {
                            i += 1;
                        }
                        float x = Transform(data[i], xScale);
                        float y = Transform(i + 1, yScale);
                        PointF p2 = new PointF(x, p.Y);
                        PointF p3 = new PointF(x, y);
                        g.DrawLine(pen, p, p2);
                        g.DrawLine(pen, p2, p3);
                        p = p3;
                    }
                    break;
                case DataMode.CumulativeFromEnd:
                    p = new PointF(Transform(data[0], xScale), Transform(data.Count, yScale));
                    for (int i = 0; i < data.Count; i++)
                    {
                        while (i + 1 < data.Count && data[i + 1] == data[i])
                        {
                            i += 1;
                        }
                        float x = Transform(data[i], xScale);
                        float y = Transform(data.Count - i - 1, yScale);
                        PointF p2 = new PointF(x, p.Y);
                        PointF p3 = new PointF(x, y);
                        g.DrawLine(pen, p, p2);
                        g.DrawLine(pen, p2, p3);
                        p = p3;
                    }
                    break;
                default:
                    throw new NotSupportedException();
            }
        }

        private static float Transform(int value, AxisScale scale)
        {
            if (scale == AxisScale.Logarithmic && value > 0)
            {
                return 1 + (float)Math.Log(value, 2);
            }
            else
            {
                return value;
            }
        }
    }
}
