#include <algorithm>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include "parsing.h"
#include "phyltree.h"

using namespace std;

struct atom_info
{
    int atom_id;
    int class_id;
    int index;
    
    atom_info(int atom_id, int class_id, int index)
    {
        this->atom_id = atom_id;
        this->class_id = class_id;
        this->index = index;
    }
};

vector<vector<atom_info>*> seqs;
vector<string> names;
int next_atom_id;
int next_class_id;

void shift(int class_id, int index, int length)
{
    for (unsigned int i = 0; i < seqs.size(); i++)
    {
        vector<atom_info> &seq = *seqs[i];
        for (unsigned int j = 0; j < seq.size(); j++)
        {
            if (seq[j].class_id == class_id)
            {
                if (seq[j].index > index)
                {
                    seq[j].index += length;
                }
                else if (seq[j].index < -index)
                {
                    seq[j].index -= length;
                }
            }
        }
    }
}

void split(int class_id, int index)
{
    for (unsigned int i = 0; i < seqs.size(); i++)
    {
        vector<atom_info> &seq = *seqs[i];
        int last_changed_atom_id = -1;
        for (unsigned int j = 0; j < seq.size(); j++)
        {
            if (seq[j].class_id == class_id && abs(seq[j].index) > index)
            {
                last_changed_atom_id = seq[j].atom_id;
                seq[j].atom_id = next_atom_id;
                seq[j].class_id = next_class_id;
                if (seq[j].index > 0)
                {
                    seq[j].index -= index;
                }
                else
                {
                    seq[j].index += index;
                }
            }
            if (last_changed_atom_id != -1 && (j + 1 == seq.size() || seq[j + 1].atom_id != last_changed_atom_id))
            {
                last_changed_atom_id = -1;
                next_atom_id += 1;
            }
        }
    }
    next_class_id += 1;
}

void split(vector<atom_info> &seq, int offset)
{
    if (offset > 0 && offset < (int)seq.size() && seq[offset - 1].atom_id == seq[offset].atom_id)
    {
        split(seq[offset].class_id, abs(seq[offset - 1].index + seq[offset].index) / 2);
    }
}

void insertion(vector<atom_info> &seq, int offset, int length, bool long_event)
{
    int atom_id;
    int class_id;
    int index;
    if (long_event)
    {
        split(seq, offset);
        atom_id = next_atom_id;
        class_id = next_class_id;
        index = 0;
        next_atom_id += 1;
        next_class_id += 1;
    }
    else
    {
        if (offset == (int)seq.size())
        {
            // We assume that the sequence is always at least 1 base long.
            atom_id = seq[offset - 1].atom_id;
            class_id = seq[offset - 1].class_id;
            index = seq[offset - 1].index;
        }
        else if (offset == 0 || seq[offset - 1].atom_id != seq[offset].atom_id)
        {
            atom_id = seq[offset].atom_id;
            class_id = seq[offset].class_id;
            index = seq[offset].index - 1;
        }
        else
        {
            atom_id = seq[offset].atom_id;
            class_id = seq[offset].class_id;
            index = (seq[offset - 1].index + seq[offset].index - 1) / 2;
        }
    }
    vector<atom_info> inseq;
    inseq.reserve(length);
    if (index >= 0)
    {
        shift(class_id, index, length);
        for (int i = 0; i < length; i++)
        {
            inseq.push_back(atom_info(atom_id, class_id, index + i + 1));
        }
    }
    else
    {
        shift(class_id, -index - 1, length);
        for (int i = -length; i < 0; i++)
        {
            inseq.push_back(atom_info(atom_id, class_id, index + i + 1));
        }
    }
    seq.insert(seq.begin() + offset, inseq.begin(), inseq.end());
}

void deletion(vector<atom_info> &seq, int offset, int length, bool long_event)
{
    if (long_event)
    {
        split(seq, offset);
        split(seq, offset + length);
    }
    seq.erase(seq.begin() + offset, seq.begin() + offset + length);
}

void duplication(vector<atom_info> &seq, int offset, int length, int source, bool inverted, bool long_event)
{
    if (long_event)
    {
        if (source < 0)
        {
            int ins_len = min(length, -source);
            insertion(seq, offset, ins_len, long_event);
            if (!inverted)
            {
                offset += ins_len;
            }
            length -= ins_len;
            source += ins_len;
        }
        if (source + length > (int)seq.size())
        {
            int ins_len = min(source + length - (int)seq.size(), length);
            insertion(seq, offset, ins_len, long_event);
            if (inverted)
            {
                offset += ins_len;
            }
            length -= ins_len;
        }
        if (length > 0)
        {
            split(seq, source);
            split(seq, source + length);
            vector<atom_info> inseq;
            inseq.reserve(length);
            if (inverted)
            {
                for (int i = source + length - 1; i >= source; i--)
                {
                    inseq.push_back(atom_info(next_atom_id, seq[i].class_id, -seq[i].index));
                    if (i == source || seq[i - 1].atom_id != seq[i].atom_id)
                    {
                        next_atom_id += 1;
                    }
                }
            }
            else
            {
                for (int i = source; i < source + length; i++)
                {
                    inseq.push_back(atom_info(next_atom_id, seq[i].class_id, seq[i].index));
                    if (i + 1 == source + length || seq[i + 1].atom_id != seq[i].atom_id)
                    {
                        next_atom_id += 1;
                    }
                }
            }
            seq.insert(seq.begin() + offset, inseq.begin(), inseq.end());
        }
    }
    else
    {
        insertion(seq, offset, length, long_event);
    }
}

vector<atom_info>* copy(vector<atom_info> &source)
{
    unsigned int start = 0;
    unsigned int pos = 0;
    vector<atom_info> &seq = *(new vector<atom_info>(source));
    while (pos < seq.size())
    {
        while (pos < seq.size() && seq[pos].atom_id == seq[start].atom_id)
        {
            pos += 1;
        }
        while (start < pos)
        {
            seq[start].atom_id = next_atom_id;
            start += 1;
        }
        next_atom_id += 1;
        start = pos;
    }
    return &seq;
}

void run(istream &input, vector<atom_info> &seq, phyltree *tree)
{
    string line;
    while (line = read_line(input), !input.eof())
    {
        if (line == ">up")
        {
            return;
        }
        else if (line == ">left")
        {
            seqs.push_back(copy(seq));
            names.push_back(tree->left->species);
            run(input, *seqs.back(), tree->left);
        }
        else if (line == ">right")
        {
            seqs.push_back(copy(seq));
            names.push_back(tree->right->species);
            run(input, *seqs.back(), tree->right);
        }
        else
        {
            bool long_event = !(line.size() > 0 && line[0] == '.');
            istringstream ss(read_line(input));
            string type;
            int offset;
            int length;
            ss >> type >> offset >> length;
            if (type == "ins")
            {
                insertion(seq, offset, length, long_event);
            }
            else if (type == "del")
            {
                deletion(seq, offset, length, long_event);
            }
            else if (type == "dup")
            {
                int source;
                char strand;
                ss >> source >> strand;
                duplication(seq, offset, length, source, strand == '+', long_event);
            }
            else
            {
                cerr << "Warning: Unknown event type '" << type << "'." << endl;
            }
        }
    }
}

void write_output()
{
    for (unsigned int i = 0; i < seqs.size(); i++)
    {
        unsigned int start = 0;
        unsigned int pos = 0;
        vector<atom_info> &seq = *seqs[i];
        while (pos < seq.size())
        {
            while (pos < seq.size() && seq[pos].atom_id == seq[start].atom_id)
            {
                pos += 1;
            }
            cout << names[i] << "\t" << seq[start].atom_id << "\t" << seq[start].class_id << "\t" << (seq[start].index > 0 ? "1" : "-1") << "\t" << start << "\t" << pos << endl;
            start = pos;
        }
    }
}

void run(istream &input)
{
    phyltree *tree = new phyltree(read_line(input));
    int root_length = read_int(input);
    read_line(input);
    seqs.push_back(new vector<atom_info>());
    names.push_back(tree->species);
    vector<atom_info> &seq = *seqs[0];
    for (int i = 0; i < root_length; i++)
    {
        seq.push_back(atom_info(0, 0, i + 1));
    }
    next_atom_id = 1;
    next_class_id = 1;
    run(input, seq, tree);
    write_output();
    delete tree;
    for (unsigned int i = 0; i < seqs.size(); i++)
    {
        delete seqs[i];
    }
}

void print_usage()
{
    cout << "Usage: logatomize [FILE]" << endl;
}

int main(int argc, char **argv)
{
    switch (argc)
    {
        case 1:
            run(cin);
            break;
        case 2:
            {
                ifstream f(argv[1]);
                run(f);
            }
            break;
        default:
            print_usage();
            break;
    }
}
