#include <cmath>
#include <sstream>
#include "distribution.h"
#include "random.h"

using namespace std;

distribution* distribution::from_string(const string &description)
{
    istringstream ss(description);
    string name;
    ss >> name;
    if (name == "EXP")
    {
        double lambda;
        ss >> lambda >> ws;
        if (ss.fail())
        {
            throw (string)"Invalid arguments, expected 'EXP <number> [<number> <number>]'.";
        }
        if (ss.eof())
        {
            return new expdist(lambda);
        }
        else
        {
            double min;
            double max;
            ss >> min >> max >> ws;
            if (ss.fail() || !ss.eof())
            {
                throw (string)"Invalid arguments, expected 'EXP <number> [<number> <number>]'.";
            }
            return new expdist(lambda, min, max);
        }
    }
    else if (name == "GEO")
    {
        double p;
        ss >> p >> ws;
        if (ss.fail())
        {
            throw (string)"Invalid arguments, expected 'GEO <number> [<number> <number>]'.";
        }
        if (ss.eof())
        {
            return new geodist(p);
        }
        else
        {
            double min;
            double max;
            ss >> min >> max >> ws;
            if (ss.fail() || !ss.eof())
            {
                throw (string)"Invalid arguments, expected 'GEO <number> [<number> <number>]'.";
            }
            return new geodist(p, min, max);
        }
    }
    else if (name == "NB")
    {
        int r;
        double p;
        ss >> r >> p >> ws;
        if (ss.fail())
        {
            throw (string)"Invalid arguments, expected 'NB <integer> <number> [<number> <number>]'.";
        }
        if (ss.eof())
        {
            return new negbindist(r, p);
        }
        else
        {
            double min;
            double max;
            ss >> min >> max >> ws;
            if (ss.fail() || !ss.eof())
            {
                throw (string)"Invalid arguments, expected 'NB <integer> <number> [<number> <number>]'.";
            }
            return new negbindist(r, p, min, max);
        }
    }
    else if (name == "POW")
    {
        double a;
        ss >> a >> ws;
        if (ss.fail())
        {
            throw (string)"Invalid arguments, expected 'POW <number> [<number> <number>]'.";
        }
        if (ss.eof())
        {
            return new zipfdist(a);
        }
        else
        {
            double min;
            double max;
            ss >> min >> max >> ws;
            if (ss.fail() || !ss.eof())
            {
                throw (string)"Invalid arguments, expected 'POW <number> [<number> <number>]'.";
            }
            return new zipfdist(a, min, max);
        }
    }
    else
    {
        throw (string)"Unknown distribution identifier '" + name + "'.";
    }
}

distribution::distribution(double min, double max)
{
    this->min = min;
    this->max = max;
}

double distribution::sample()
{
    double value;
    do
    {
        value = next();
    }
    while (!(value >= min) || !(value <= max));
    return value;
}

int distribution::sample_int()
{
    return (int)round(sample());
}

constdist::constdist(double value) : distribution(value, value)
{
    this->value = value;
}

distribution* constdist::clone() const
{
    return new constdist(*this);
}

double constdist::next()
{
    return value;
}

expdist::expdist(double lambda) : distribution(0, INFINITY)
{
    init(lambda);
}

expdist::expdist(double lambda, double min, double max) : distribution(min, max)
{
    init(lambda);
}

distribution* expdist::clone() const
{
    return new expdist(*this);
}

void expdist::init(double lambda)
{
    if (!(lambda > 0))
    {
        throw (string)"Invalid argument, lambda must be positive.";
    }
    this->lambda = lambda;
}

double expdist::next()
{
    return -log(random::next_strict()) / lambda;
}

geodist::geodist(double p) : distribution(0, INFINITY)
{
    init(p);
}

geodist::geodist(double p, double min, double max) : distribution(min, max)
{
    init(p);
}

distribution* geodist::clone() const
{
    return new geodist(*this);
}

void geodist::init(double p)
{
    if (!(p > 0) || !(p <= 1))
    {
        throw (string)"Invalid argument, p must be from (0, 1].";
    }
    this->log_q = std::min(log(1 - p), -1e-300);
}

double geodist::next()
{
    return floor(log(random::next_strict()) / log_q);
}

negbindist::negbindist(int r, double p) : distribution(0, INFINITY), g(p)
{
    init(r);
}

negbindist::negbindist(int r, double p, double min, double max) : distribution(min, max), g(p)
{
    init(r);
}

distribution* negbindist::clone() const
{
    return new negbindist(*this);
}

void negbindist::init(double r)
{
    if (!(r >= 0))
    {
        throw (string)"Invalid argument, r must be non-negative.";
    }
    this->r = r;
}

double negbindist::next()
{
    double sum = 0;
    for (int i = 0; i < r; i++)
    {
        sum += g.sample();
    }
    return sum;
}

zipfdist::zipfdist(double a) : distribution(0, INFINITY)
{
    init(a);
}

zipfdist::zipfdist(double a, double min, double max) : distribution(min, max)
{
    init(a);
}

distribution* zipfdist::clone() const
{
    return new zipfdist(*this);
}

void zipfdist::init(double a)
{
    if (!(a > 1))
    {
        throw (string)"Invalid argument, a must be greater than 1.";
    }
    this->a = a;
    this->b = pow(2, a - 1);
    this->c = -1 / (a - 1);
}

// Devroye Luc (1986) Non-uniform random variate generation.
// Springer-Verlag: Berlin. p551
double zipfdist::next()
{
    double x, t;
    do
    {
        x = floor(pow(random::next_strict(), c));
        t = pow(1 + (1 / x), a - 1);
    }
    while (!(random::next() * x * (t - 1) * b <= t * (b - 1)));
    return x;
}
