#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include "distribution.h"
#include "parsing.h"
#include "random.h"
#include "simulator.h"
#include "submodel.h"

using namespace std;

// Output stream that discards all the data (like /dev/null).
class nullstream : public ostream
{
    public:
        nullstream() : ostream(NULL)
        {
        }
};

simulator sim;
ostream *output = NULL;
ostream *logfile = NULL;

void read_settings(istream &input)
{
    bool randseed_set = false;
    bool submodel_set = false;
    bool basefreq_set = false;
    bool tree_set = false;
    bool rootlen_set = false;
    bool output_set = false;
    bool logfile_set = false;
    double base_freqs[4];
    string setting;
    string value;
    while (parse_line(input, setting, value))
    {
        istringstream values(value);
        try
        {
            if (setting == "randseed")
            {
                if (randseed_set)
                {
                    throw (string)"Random seed was specified multiple times.";
                }
                unsigned int seed;
                values >> seed;
                if (values.fail() || !values.eof())
                {
                    throw (string)"Invalid syntax, expected 'randseed <integer>'.";
                }
                else
                {
                    random::seed(seed);
                    randseed_set = true;
                }
            }
            else if (setting == "submodel")
            {
                if (submodel_set)
                {
                    throw (string)"Nucleotide substitution model was specified multiple times.";
                }
                sim.substitution_model = new submodel(value);
                submodel_set = true;
            }
            else if (setting == "basefreq")
            {
                if (basefreq_set)
                {
                    throw (string)"Stationary base frequencies were specified multiple times.";
                }
                values >> base_freqs[0] >> base_freqs[1] >> base_freqs[2] >> base_freqs[3];
                if (values.fail() || !values.eof())
                {
                    throw (string)"Invalid syntax, expected 'basefreq <number> <number> <number> <number>'.";
                }
                else
                {
                    basefreq_set = true;
                }
            }
            else if (setting == "tree")
            {
                if (tree_set)
                {
                    throw (string)"Phylogenetic tree was specified multiple times.";
                }
                sim.phylogenetic_tree = new phyltree(value);
                tree_set = true;
            }
            else if (setting == "rootlen")
            {
                if (rootlen_set)
                {
                    throw (string)"Root sequence length was specified multiple times.";
                }
                double rootlen;
                values >> rootlen;
                if (values.fail() || !values.eof())
                {
                    throw (string)"Invalid syntax, expected 'rootlen <integer>'.";
                }
                else
                {
                    sim.root_length = rootlen;
                    rootlen_set = true;
                }
            }
            else if (setting == "output")
            {
                if (output_set)
                {
                    throw (string)"Output file was specified multiple times.";
                }
                output = new ofstream(value.c_str());
                output_set = true;
            }
            else if (setting == "logfile")
            {
                if (logfile_set)
                {
                    throw (string)"Logging file was specified multiple times.";
                }
                logfile = new ofstream(value.c_str());
                logfile_set = true;
            }
            else if (setting == "event")
            {
                try
                {
                    if (value != "")
                    {
                        throw (string)"Invalid syntax, nothing more expected after 'event'.";
                    }
                    ostringstream ss;
                    while (parse_line(input, setting, value) && setting != "endevent")
                    {
                        ss << setting << " " << value << endl;
                    }
                    if (setting != "endevent")
                    {
                        throw (string)"Invalid syntax, 'event' must have a matching 'endevent'.";
                    }
                    if (value != "")
                    {
                        throw (string)"Invalid syntax, nothing more expected after 'endevent'.";
                    }
                    sim.sequence_events.push_back(seqevent::from_string(ss.str()));
                }
                catch (string ex)
                {
                    cerr << "Warning: " + ex + " Event will be ignored." << endl;
                }
            }
            else
            {
                throw (string)"Unknown setting '" + setting + "'.";
            }
        }
        catch (string ex)
        {
            cerr << "Warning: " + ex + " Setting will be ignored." << endl;
            cerr.flush();
        }
    }
    if (!randseed_set)
    {
        random::seed(time(NULL));
    }
    if (submodel_set && basefreq_set)
    {
        sim.substitution_model->set_base_freqs(base_freqs[0], base_freqs[1], base_freqs[2], base_freqs[3]);
    }
    if (!output_set)
    {
        output = &cout;
    }
    if (!logfile_set)
    {
        logfile = new nullstream();
    }
}

void print_usage()
{
    cout << "Usage: simseq [FILE]" << endl;
}

int main(int argc, char **argv)
{
    switch (argc)
    {
        case 1:
            // Read settings from the standard input.
            read_settings(cin);
            break;
        case 2:
            // Read settings from file specified by the first argument.
            {
                ifstream f(argv[1]);
                read_settings(f);
            }
            break;
        default:
            // Invalid number of arguments, print info message.
            print_usage();
            return 0;
    }
    try
    {
        sim.run(*output, *logfile);
    }
    catch (string ex)
    {
        cerr << endl << "*** Error: " + ex + " Simulation aborted. ***" << endl;
    }
    if (output != &cout)
    {
        output->flush();
        delete output;
    }
    logfile->flush();
    delete logfile;
    return 0;
}
