#include <cmath>
#include <iostream>
#include <sstream>

#include "parsing.h"
#include "random.h"
#include "seqevent.h"
#include "sequence.h"

using namespace std;

seqevent* seqevent::from_string(const string &description)
{
    string name = "";
    string type = "";
    double rate = NAN;
    distribution *length = NULL;
    distribution *dist = NULL;
    double inv_rate = NAN;
    istringstream ss(description);
    string setting;
    string value;
    while (parse_line(ss, setting, value))
    {
        istringstream values(value);
        try
        {
            if (setting == "name")
            {
                if (name != "")
                {
                    throw (string)"Event name was specified multiple times.";
                }
                if (value == "")
                {
                    throw (string)"Invalid syntax, expected 'name <string>'.";
                }
                name = value;
            }
            else if (setting == "type")
            {
                if (type != "")
                {
                    throw (string)"Event type was specified multiple times.";
                }
                if (value == "")
                {
                    throw (string)"Invalid syntax, expected 'type <string>'.";
                }
                type = value;
            }
            else if (setting == "rate")
            {
                if (!isnan(rate))
                {
                    throw (string)"Event rate was specified multiple times.";
                }
                values >> rate;
                if (values.fail() || !values.eof())
                {
                    throw (string)"Invalid syntax, expected 'rate <number>'.";
                }
            }
            else if (setting == "length")
            {
                if (length != NULL)
                {
                    throw (string)"Event length was specified multiple times.";
                }
                length = distribution::from_string(value);
            }
            else if (setting == "dist")
            {
                if (dist != NULL)
                {
                    throw (string)"Event distance was specified multiple times.";
                }
                dist = distribution::from_string(value);
            }
            else if (setting == "inv")
            {
                if (!isnan(inv_rate))
                {
                    throw (string)"Event inversion rate was specified multiple times.";
                }
                double inv;
                values >> inv;
                if (values.fail() || !values.eof())
                {
                    throw (string)"Invalid syntax, expected 'inv <number>'.";
                }
                else
                {
                    inv_rate = inv;
                }
            }
            else
            {
                throw (string)"Unknown setting '" + setting + "'.";
            }
        }
        catch (string ex)
        {
            cerr << "Warning: " + ex + " Setting will be ignored." << endl;
        }
    }
    seqevent *res = NULL;
    string ex;
    if (isnan(rate))
    {
        ex = "Event rate not specified.";
    }
    else if (!(rate > 0))
    {
        ex = "Event rate must be positive.";
    }
    else if (length == NULL)
    {
        ex = "Event length distribution not specified.";
    }
    else if (type == "insertion")
    {
        res = new insertion(name, rate, *length);
    }
    else if (type == "deletion")
    {
        res = new deletion(name, rate, *length);
    }
    else if (type == "duplication")
    {
        if (isnan(inv_rate))
        {
            inv_rate = 0.5;
        }
        if (dist == NULL)
        {
            ex = "Duplication distance distribution not specified.";
        }
        else
        {
            res = new duplication(name, rate, *length, *dist, inv_rate);
        }
    }
    else if (type == "")
    {
        ex = "Event type not specified.";
    }
    else
    {
        ex = "Unknown event type '" + type + "'.";
    }
    if (length != NULL)
    {
        delete length;
    }
    if (dist != NULL)
    {
        delete dist;
    }
    if (res == NULL)
    {
        throw ex;
    }
    return res;
}

seqevent::seqevent(const string &name, double rate) : name(name), time(rate)
{
    next_time = time.sample();
}

seqevent::~seqevent()
{
}

double seqevent::next_occurence(int seqlen)
{
    return next_time / affecting_positions(seqlen);
}

void seqevent::simulate(vector<char> &seq, double base_freqs[4], ostream &logfile)
{
    next_time = time.sample();
    logfile << name << endl;
}

void seqevent::wait(double time, int seqlen)
{
    next_time -= time * affecting_positions(seqlen);
}

insertion::insertion(const string &name, double rate, const distribution &length) : seqevent(name, rate)
{
    this->length = length.clone();
}

insertion::insertion(const insertion &ins) : seqevent(ins)
{
    length = ins.length->clone();
}

insertion& insertion::operator=(const insertion &ins)
{
    if (this != &ins)
    {
        name = ins.name;
        time = ins.time;
        next_time = ins.next_time;
        delete length;
        length = ins.length->clone();
    }
    return *this;
}

insertion::~insertion()
{
    delete length;
}

seqevent* insertion::clone() const
{
    return new insertion(*this);
}

bool insertion::needs_substitutions()
{
    return false;
}

int insertion::affecting_positions(int seqlen)
{
    return seqlen + 1;
}

void insertion::simulate(vector<char> &seq, double base_freqs[4], ostream &logfile)
{
    seqevent::simulate(seq, base_freqs, logfile);
    int offset = random::next_int(seq.size() + 1);
    int len = length->sample_int();
    vector<char> inseq;
    rand_seq(len, inseq, base_freqs);
    seq.insert(seq.begin() + offset, inseq.begin(), inseq.end());
    logfile << "ins " << offset << " " << len << endl;
}

deletion::deletion(const string &name, double rate, const distribution &length) : seqevent(name, rate)
{
    this->length = length.clone();
    next_length = this->length->sample_int();
}

deletion::deletion(const deletion &del) : seqevent(del)
{
    length = del.length->clone();
    next_length = del.next_length;
}

deletion& deletion::operator=(const deletion &del)
{
    if (this != &del)
    {
        name = del.name;
        time = del.time;
        next_time = del.next_time;
        delete length;
        length = del.length->clone();
        next_length = del.next_length;
    }
    return *this;
}

deletion::~deletion()
{
    delete length;
}

seqevent* deletion::clone() const
{
    return new deletion(*this);
}

bool deletion::needs_substitutions()
{
    return false;
}

int deletion::affecting_positions(int seqlen)
{
    return seqlen + next_length;
}

void deletion::simulate(vector<char> &seq, double base_freqs[4], ostream &logfile)
{
    seqevent::simulate(seq, base_freqs, logfile);
    int offset = random::next_int(seq.size() + next_length) - next_length;
    int len = next_length;
    next_length = length->sample_int();
    if (offset < 0)
    {
        len += offset;
        offset = 0;
    }
    if (offset + len > (int)seq.size())
    {
        len = (int)seq.size() - offset;
    }
    seq.erase(seq.begin() + offset, seq.begin() + offset + len);
    logfile << "del " << offset << " " << len << endl;
}

duplication::duplication(const string &name, double rate, const distribution &length, const distribution &distance, double inv_rate) : seqevent(name, rate)
{
    this->length = length.clone();
    this->dist = distance.clone();
    this->inv_rate = inv_rate;
}

duplication::duplication(const duplication &dup) : seqevent(dup)
{
    length = dup.length->clone();
    dist = dup.dist->clone();
    inv_rate = dup.inv_rate;
}

duplication& duplication::operator=(const duplication &dup)
{
    if (this != &dup)
    {
        name = dup.name;
        time = dup.time;
        next_time = dup.next_time;
        delete length;
        length = dup.length->clone();
        delete dist;
        dist = dup.dist->clone();
        inv_rate = dup.inv_rate;
    }
    return *this;
}

duplication::~duplication()
{
    delete length;
    delete dist;
}

seqevent* duplication::clone() const
{
    return new duplication(*this);
}

bool duplication::needs_substitutions()
{
    return true;
}

int duplication::affecting_positions(int seqlen)
{
    return seqlen + 1;
}

void duplication::simulate(vector<char> &seq, double base_freqs[4], ostream &logfile)
{
    seqevent::simulate(seq, base_freqs, logfile);
    int offset = random::next_int(seq.size() + 1);
    int len = length->sample_int();
    bool inv = (random::next() < inv_rate);
    int source = offset;
    if (random::next_bool())
    {
        source += dist->sample_int();
    }
    else
    {
        source -= dist->sample_int() + len;
    }
    vector<char> inseq;
    inseq.reserve(len);
    if (inv)
    {
        for (int i = source + len - 1; i >= source; i--)
        {
            if (i >= 0 && i < (int)seq.size())
            {
                inseq.push_back(compl_base(seq[i]));
            }
            else
            {
                inseq.push_back(rand_base(base_freqs));
            }
        }
    }
    else
    {
        for (int i = source; i < source + len; i++)
        {
            if (i >= 0 && i < (int)seq.size())
            {
                inseq.push_back(seq[i]);
            }
            else
            {
                inseq.push_back(rand_base(base_freqs));
            }
        }
    }
    seq.insert(seq.begin() + offset, inseq.begin(), inseq.end());
    logfile << "dup " << offset << " " << len << " " << source << " " << (inv ? "-" : "+") << endl;
}
