#include <algorithm>
#include <cmath>
#include "sequence.h"
#include "simulator.h"

using namespace std;

simulator::simulator()
{
    substitution_model = NULL;
    phylogenetic_tree = NULL;
    root_length = 0;
}

// Creates a deep copy of the simulator.
simulator::simulator(const simulator &s) : sequence_events(s.sequence_events)
{
    if (s.substitution_model != NULL)
    {
        substitution_model = new submodel(*s.substitution_model);
    }
    else
    {
        substitution_model = NULL;
    }
    if (s.phylogenetic_tree != NULL)
    {
        phylogenetic_tree = new phyltree(*s.phylogenetic_tree);
    }
    else
    {
        phylogenetic_tree = NULL;
    }
    root_length = s.root_length;
}

// Creates a deep copy of the simulator.
simulator& simulator::operator=(const simulator &s)
{
    if (this != &s)
    {
        cleanup();
        if (s.substitution_model != NULL)
        {
            substitution_model = new submodel(*s.substitution_model);
        }
        else
        {
            substitution_model = NULL;
        }
        if (s.phylogenetic_tree != NULL)
        {
            phylogenetic_tree = new phyltree(*s.phylogenetic_tree);
        }
        else
        {
            phylogenetic_tree = NULL;
        }
        root_length = s.root_length;
        for (unsigned int i = 0; i < s.sequence_events.size(); i++)
        {
            sequence_events.push_back(s.sequence_events[i]->clone());
        }
    }
    return *this;
}

simulator::~simulator()
{
    cleanup();
}

void simulator::cleanup()
{
    if (substitution_model != NULL)
    {
        delete substitution_model;
    }
    if (phylogenetic_tree != NULL)
    {
        delete phylogenetic_tree;
    }
    for (unsigned int i = 0; i < sequence_events.size(); i++)
    {
        delete sequence_events[i];
    }
    sequence_events.clear();
}

void simulator::validate()
{
    if (substitution_model == NULL)
    {
        throw (string)"Nucleotide substitution model not specified.";
    }
    if (phylogenetic_tree == NULL)
    {
        throw (string)"Phylogenetic tree not specified.";
    }
    if (!(root_length >= 0))
    {
        throw (string)"Root sequence length must be non-negative.";
    }
    substitution_model->get_base_freqs(base_freqs[0], base_freqs[1], base_freqs[2], base_freqs[3]);
    
}

// Evolves the given sequence for a given amount of time.
void simulator::evolve(vector<char> &seq, double t, ostream &logfile)
{
    // How long ago we last simulated the substitution model.
    double subst_t = 0;
    
    // The main loop which runs until the simulation time t is exhausted.
    while (true)
    {
        ////////////////////////////////////////////////////////////////////////
        // Determine which will be the next event.                            //
        ////////////////////////////////////////////////////////////////////////
        
        // Value of -1 means that the next event will be the speciation.
        int next_event = -1;
        double dt = t;
        for (unsigned int i = 0; i < sequence_events.size(); i++)
        {
            double next_time = sequence_events[i]->next_occurence(seq.size());
            if (next_time < dt)
            {
                next_event = i;
                dt = next_time;
            }
        }
        
        ////////////////////////////////////////////////////////////////////////
        // Advance the simulation time. The calls to wait() are necessary for //
        // proper functioning of the seqevent objects.                        //
        ////////////////////////////////////////////////////////////////////////
        
        subst_t += dt;
        t -= dt;
        for (unsigned int i = 0; i < sequence_events.size(); i++)
        {
            sequence_events[i]->wait(dt, seq.size());
        }
        
        ////////////////////////////////////////////////////////////////////////
        // Simulate the substitution model if necessary.                      //
        ////////////////////////////////////////////////////////////////////////
        
        if (next_event == -1 || sequence_events[next_event]->needs_substitutions())
        {
            double pmat[4][4];
            substitution_model->make_pmatrix(subst_t, pmat);
            subst_t = 0;
            subst(seq, pmat);
        }
        
        ////////////////////////////////////////////////////////////////////////
        // Simulate the event.                                                //
        ////////////////////////////////////////////////////////////////////////
        
        if (next_event == -1)
        {
            break;
        }
        else
        {
            sequence_events[next_event]->simulate(seq, base_freqs, logfile);
        }
    }
}

void simulator::run(vector<char> &seq, phyltree *tree, ostream &output, ostream &logfile)
{
    write_fasta(output, seq, tree->species);
    if (tree->left != NULL)
    {
        logfile << ">left" << endl;
        vector<char> seq_copy(seq);
        evolve(seq_copy, tree->left_length, logfile);
        run(seq_copy, tree->left, output, logfile);
        logfile << ">up" << endl;
    }
    if (tree->right != NULL)
    {
        logfile << ">right" << endl;
        vector<char> seq_copy(seq);
        evolve(seq_copy, tree->right_length, logfile);
        run(seq_copy, tree->right, output, logfile);
        logfile << ">up" << endl;
    }
}

void simulator::run(ostream &output, ostream &logfile)
{
    validate();
    vector<char> seq;
    rand_seq(root_length, seq, base_freqs);
    logfile << phylogenetic_tree->to_string() << endl;
    logfile << root_length << endl;
    run(seq, phylogenetic_tree, output, logfile);
}
