#include <sstream>
#include <vector>
#include "submodel.h"

using namespace std;

// Parses a give istream into a list of doubles.
void read_doubles(istream &input, vector<double> &values)
{
    input >> ws;
    while (!input.eof())
    {
        double d;
        input >> d >> ws;
        if (input.fail())
        {
            throw (string)"Expected numeric parameters to the substitution model.";
        }
        else if (!(d >= 0))
        {
            throw (string)"Parameters to substitution model must be non-negative.";
        }
        values.push_back(d);
    }
}

submodel::submodel(const string &description)
{
    ////////////////////////////////////////////////////////////////////////////
    // Parse the description string. Expected format is <name> [<params>].    //
    ////////////////////////////////////////////////////////////////////////////
    
    istringstream ss(description);
    string name;
    vector<double> params;
    ss >> name;
    read_doubles(ss, params);
    
    ////////////////////////////////////////////////////////////////////////////
    // Set default values for private fields.                                 //
    ////////////////////////////////////////////////////////////////////////////
    
    a = b = c = d = e = f = 1;
    f84 = false;
    needs_freqs = true;
    
    ////////////////////////////////////////////////////////////////////////////
    // Change the values according to the specified model and its parameters. //
    ////////////////////////////////////////////////////////////////////////////
    
    if (name == "JC" || name == "F81")
    {
        if (params.size() != 0)
        {
            throw (string)"Model " + name + " expects no paramaters.";
        }
    }
    else if (name == "K80" || name == "HKY")
    {
        if (params.size() < 1 || params.size() > 2)
        {
            throw (string)"Model " + name + " expects one or two paramaters.";
        }
        a = f = params[0];
        if (params.size() == 2)
        {
            b = c = d = e = params[1];
        }
    }
    else if (name == "TrNef" || name == "TrN")
    {
        if (params.size() < 2 || params.size() > 3)
        {
            throw (string)"Model " + name + " expects two or three paramaters.";
        }
        a = params[0];
        f = params[1];
        if (params.size() == 3)
        {
            b = c = d = e = params[2];
        }
    }
    else if (name == "K81" || name == "K81uf")
    {
        if (params.size() < 2 || params.size() > 3)
        {
            throw (string)"Model " + name + " expects two or three paramaters.";
        }
        b = e = params[0];
        c = d = params[1];
        if (params.size() == 3)
        {
            a = f = params[2];
        }
    }
    else if (name == "TIMef" || name == "TIM")
    {
        if (params.size() < 3 || params.size() > 4)
        {
            throw (string)"Model " + name + " expects three or four paramaters.";
        }
        a = params[0];
        b = e = params[1];
        c = d = params[2];
        if (params.size() == 4)
        {
            f = params[3];
        }
    }
    else if (name == "TVMef" || name == "TVM")
    {
        if (params.size() < 4 || params.size() > 5)
        {
            throw (string)"Model " + name + " expects four or five paramaters.";
        }
        b = params[0];
        c = params[1];
        d = params[2];
        e = params[3];
        if (params.size() == 5)
        {
            a = f = params[4];
        }
    }
    else if (name == "SYM" || name == "GTR")
    {
        if (params.size() < 5 || params.size() > 6)
        {
            throw (string)"Model " + name + " expects five or six paramaters.";
        }
        a = params[0];
        b = params[1];
        c = params[2];
        d = params[3];
        e = params[4];
        if (params.size() == 6)
        {
            f = params[5];
        }
    }
    else if (name == "F84ef" || name == "F84")
    {
        if (params.size() < 1 || params.size() > 2)
        {
            throw (string)"Model " + name + " expects one or two paramaters.";
        }
        f84 = true;
        a = params[0];
        if (params.size() == 2)
        {
            b = c = d = e = params[1];
        }
    }
    else
    {
        throw (string)"Unknown model identifier '" + name + "'.";
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // These models use the default stationary base frequencies.              //
    ////////////////////////////////////////////////////////////////////////////
    
    if (name == "JC" || name == "K80" || name == "TrNef" || name == "K81" || name == "TIMef" || name == "TVMef" || name == "SYM" || name == "F84ef")
    {
        set_base_freqs(0.25, 0.25, 0.25, 0.25);
    }
}

// Returns true if the stationary base frequencies have not been specified yet.
bool submodel::needs_base_freqs()
{
    return needs_freqs;
}

void submodel::get_base_freqs(double &pt, double &pc, double &pa, double &pg)
{
    if (needs_freqs)
    {
        throw (string)"Stationary base frequencies were not specified.";
    }
    pt = base_freqs[0];
    pc = base_freqs[1];
    pa = base_freqs[2];
    pg = base_freqs[3];
}

// Scales the values so that they sum to one.
void normalize(double &pt, double &pc, double &pa, double &pg)
{
    if (!(pt >= 0) || !(pc >= 0) || !(pa >= 0) || !(pg >= 0))
    {
        throw (string)"Stationary base frequencies must be non-negative.";
    }
    double sum = pt + pc + pa + pg;
    if (sum == 0)
    {
        throw (string)"Stationary base frequencies cannot be all zero.";
    }
    pt /= sum;
    pc /= sum;
    pa /= sum;
    pg /= sum;
}

void submodel::set_base_freqs(double pt, double pc, double pa, double pg)
{
    if (!needs_base_freqs())
    {
        throw (string)"Stationary base frequencies were not expected for this model.";
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // Normalize the frequencies and save them.                               //
    ////////////////////////////////////////////////////////////////////////////
    
    normalize(pt, pc, pa, pg);
    base_freqs[0] = pt;
    base_freqs[1] = pc;
    base_freqs[2] = pa;
    base_freqs[3] = pg;
    needs_freqs = false;
    
    ////////////////////////////////////////////////////////////////////////////
    // F84 parameters depend on the values of stationary frequencies so they  //
    // couldn't be calculated during the initialization.                      //
    ////////////////////////////////////////////////////////////////////////////
    
    if (f84)
    {
        double kappa = a;
        a = (1 + kappa / (pt + pc)) * b;
        f = (1 + kappa / (pa + pg)) * b;
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // Initialize the substitution matrix from model parameters.              //
    ////////////////////////////////////////////////////////////////////////////
    
    qmat[0][0] = 0; qmat[0][1] = a; qmat[0][2] = b; qmat[0][3] = c;
    qmat[1][0] = a; qmat[1][1] = 0; qmat[1][2] = d; qmat[1][3] = e;
    qmat[2][0] = b; qmat[2][1] = d; qmat[2][2] = 0; qmat[2][3] = f;
    qmat[3][0] = c; qmat[3][1] = e; qmat[3][2] = f; qmat[3][3] = 0;
    
    ////////////////////////////////////////////////////////////////////////////
    // Multiply matrix entries by the stationary base frequencies.            //
    ////////////////////////////////////////////////////////////////////////////
    
    for (int i = 0; i < 4; i++)
    {
        for (int j = 0; j < 4; j++)
        {
            qmat[i][j] *= base_freqs[j];
        }
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // Scale the matrix to get the expected number of substitutions per site  //
    // per unit of time equal one.                                            //
    ////////////////////////////////////////////////////////////////////////////
    
    double scale = 0;
    for (int i = 0; i < 4; i++)
    {
        for (int j = 0; j < 4; j++)
        {
            scale += base_freqs[i] * qmat[i][j];
        }
    }
    if (scale > 0)
    {
        for (int i = 0; i < 4; i++)
        {
            for (int j = 0; j < 4; j++)
            {
                qmat[i][j] /= scale;
            }
        }
    }
    else
    {
        // All entries are zero so we can't scale them.
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // Initialize the diagonal so that all rows sum to zero.                  //
    ////////////////////////////////////////////////////////////////////////////
    
    for (int i = 0; i < 4; i++)
    {
        double sum = 0;
        for (int j = 0; j < 4; j++)
        {
            sum += qmat[i][j];
        }
        qmat[i][i] = -sum;
    }
}

// Exponentiates the Q matrix to obtain P matrix for a given time length.
void submodel::make_pmatrix(double time, double pmat[4][4])
{
    if (needs_freqs)
    {
        throw (string)"Stationary base frequencies were not specified.";
    }
    pmat_qrev(qmat, base_freqs, time, pmat);
}
